<?php

class Truevo_Widget_Configuration_Builder
{
    /**
     * The order to build the widget configuration for.
     *
     * @var WC_Order
     */
    private $_order;

    /**
     * The configuration settings saved in the admin for the plugin.
     *
     * @var mixed
     */
    private $_gateway_settings;

    /**
     * Constructor
     *
     * @param WC_Order $order
     */
    public function __construct( WC_Order $order = null )
    {
        $this->_order = $order;
        $this->_gateway_settings = get_option('woocommerce_truevo_gateway_settings');
    }

    /**
     * Get either the Order ID or the cart hash.
     *
     * @return string
     */
    public function getOrderIdentifierAttribute()
    {
        return (string)(!is_null($this->_order) ? $this->_order->get_id() : WC()->cart->get_cart_hash());
    }

    /**
     * Return the type of transaction.
     *
     * @return string
     */
    public function getTransactionTypeAttribute()
    {
        return ($this->_gateway_settings['enable_authorization'] === 'yes') ? 'authorize' : 'sale';
    }

    /**
     * Return the widget configuration (ie. theming, style).
     *
     * @return array
     */
    public function getWidgetConfigurationAttribute()
    {
        $locale = defined('ICL_LANGUAGE_CODE') ? ICL_LANGUAGE_CODE : get_locale();
        $locale_array = explode("_", $locale);
        $language = isset($locale_array[0]) ? $locale_array[0] : 'EN';

        return [
            'language' => strtoupper($language),
            'showTruevo' => $this->_gateway_settings['show_logo'] === 'yes',
            'iframeStyles' => [
                'theme' => $this->_gateway_settings['widget_theme'],
                'isDarkTheme' => $this->_gateway_settings['dark_mode'] === 'yes',
                'style' => $this->minifyCss($this->_gateway_settings['css_styles'])
            ],
            'isInline' => truevo_is_inline_checkout_flow()
        ];
    }

    /**
     * Clean up the CSS for the widget configuration.
     *
     * @return string
     */
    private function minifyCss($css)
    {
        return str_replace(array("\r", "\n"), '', html_entity_decode($css));
    }

    /**
     * Return the transaction details for the widget.
     *
     * @return array
     */
    public function getTransactionAttribute()
    {
        $total_amount = !is_null($this->_order) ? $this->_order->get_total() : WC()->cart->total;
        $total_amount_formatted = number_format($total_amount, 2, '.', '');

        return [
            'paymentReference' => $this->_gateway_settings['store_name'],
            'merchantReference' => $this->getOrderIdentifierAttribute(),
            'amount' => [
                'totalAmount' => $total_amount_formatted,
                'currencyAlphaCode' => get_woocommerce_currency(),
                'details' => [
                    'paymentAmount' => $total_amount_formatted,
                    'gratuityAmount' => 0
                ]
            ],
            'source' => 'ECOMMERCE'
        ];
    }

    /**
     * Return the merchant details for the widget.
     *
     * @return array
     */
    public function getMerchantAttribute()
    {
        return [
            'descriptor' => $this->_gateway_settings['store_name'],
            'mid' => truevo_get_mid(),
            'tid' => truevo_get_tid()
        ];
    }

    /**
     * Return the customer details for the attribute.
     *
     * @return array
     */
    public function getCustomerAttribute()
    {
        $customer_object = !is_null($this->_order) ? $this->_order : WC()->customer;

        return [
            'firstName' => $customer_object->get_billing_first_name(),
            'lastName' => $customer_object->get_billing_last_name(),
            'contact' => [
                'email' => $customer_object->get_billing_email(),
                'ip' => truevo_get_client_ip()
            ],
            'physicalAddress' => [
                'street1' => $customer_object->get_shipping_address_1(),
                'street2' => $customer_object->get_shipping_address_2(),
                'city' => $customer_object->get_shipping_city(),
                'state' => $customer_object->get_shipping_state(),
                'postCodeZip' => $customer_object->get_shipping_postcode(),
                'countryAlphaCode' => $customer_object->get_shipping_country(),
            ],
            'billingAddress' => [
                'street1' => $customer_object->get_billing_address_1(),
                'street2' => $customer_object->get_billing_address_2(),
                'city' => $customer_object->get_billing_city(),
                'state' => $customer_object->get_billing_state(),
                'postCodeZip' => $customer_object->get_billing_postcode(),
                'countryAlphaCode' => $customer_object->get_billing_country(),
            ]
        ];
    }

    /**
     * NOT YET IMPLEMENTED.
     * Return any additional information for the transaction.
     *
     * @return array
     */
    public function getAdditionalDataAttribute()
    {
        return [

            [
            'key' => 'transactionSource',
            'value' => 'PLG'
            ]
        ];
    }

    /**
     * Return the tokenization details for a transaction.
     *
     * @return array
     */
    public function getTokenizationAttribute()
    {
        return [
            'merchantShopperReference' => truevo_get_merchant_shopper_reference(),
            'tokenRegistrationProcess' => 'Optional'
        ];
    }

    /**
     * Check if tokenization should be used for the transaction.
     *
     * @return bool
     */
    public function canAddTokenizationAttribute()
    {
        return ($this->_gateway_settings['enable_tokenization'] == 'yes' && truevo_get_merchant_shopper_reference() !== null);
    }

    /**
     * Return the fully built array for the widget configuration.
     *
     * @return array
     */
    public function build()
    {
        // Instantiate a blank configuration variable
        $configuration = [];

        // Transaction type
        $configuration['transactionType'] = $this->getTransactionTypeAttribute();

        // Widget configuration
        $configuration['widgetConfiguration'] = $this->getWidgetConfigurationAttribute();

        // Transaction
        $configuration['transaction'] = $this->getTransactionAttribute();

        // Merchant
        $configuration['merchant'] = $this->getMerchantAttribute();

        // Customer
        $configuration['customer'] = $this->getCustomerAttribute();

        // Additional data
        $configuration['additionalData'] = $this->getAdditionalDataAttribute();

        // Tokenization
        if ($this->canAddTokenizationAttribute()) {
            $configuration['tokenization'] = $this->getTokenizationAttribute();
        }

        // Return configuration
        return $configuration;
    }
}