<?php
/*
 * Plugin Name: Truevo Payments
 * Plugin URI: https://truevo.com/
 * Description: Accept online payments with the Truevo Payment Plugin. Let your customers save their card details for repeat purchases. Perform reversals, refunds, and captures. PCI compliant and 3DS v2 enabled to keep transactions secure.
 * Version: 1.9.14
 * Author: Truevo Payments
 * Author URI: https://truevo.com/
 * Text Domain: truevo-gateway-woocommerce
 * Domain Path: /languages
 */

// Make sure WooCommerce is active
if (!in_array('woocommerce/woocommerce.php', apply_filters('active_plugins', get_option('active_plugins')))) {
    return;
}

include('includes/truevo-gateway-api.php');
include('includes/truevo-widget-configuration-builder.php');

if (!function_exists('truevo_add_gateway_class')) {
    add_filter('woocommerce_payment_gateways', 'truevo_add_gateway_class');

    /**
     * Add the Truevo Payment Gateway to the list of active gateways.
     *
     * @access public
     * @param array $gateways
     * @return array
     */
    function truevo_add_gateway_class($gateways)
    {
        $gateways[] = 'WC_Truevo_Gateway';

        return $gateways;
    }
}

if (!function_exists('truevo_gateway_action_links')) {
    add_filter('plugin_action_links_' . plugin_basename(__FILE__), 'truevo_gateway_action_links');

    /**
     * Add the settings link for the Truevo Payment Gateway to WooCommerce.
     *
     * @access public
     * @param mixed $links
     * @return array
     */
    function truevo_gateway_action_links($links)
    {
        $plugin_links = array();

        if (function_exists('WC')) {
            if (version_compare(WC()->version, '2.6', '>=')) {
                $section_slug = 'truevo_gateway';
            } else {
                $section_slug = strtolower('WC_Truevo_Gateway');
            }
            $setting_url = admin_url('admin.php?page=wc-settings&tab=checkout&section=' . $section_slug);
            $plugin_links[] = '<a href="' . esc_url($setting_url) . '">' . esc_html__('Settings', 'truevo-gateway-woocommerce') . '</a>';
        }

        return array_merge($plugin_links, $links);
    }
}

// Initialise the Truevo Payment Gateway
add_action('plugins_loaded', 'truevo_init_gateway_class');

/**
 * Add in the WC_Truevo_Gateway class.
 *
 * @access public
 * @return void
 */
function truevo_init_gateway_class()
{
    class WC_Truevo_Gateway extends WC_Payment_Gateway
    {
        /**
         * Class constructor
         */
        public function __construct()
        {
            $this->id = 'truevo_gateway';
            $this->has_fields = true;
            $this->method_title = 'Truevo Payments';
            $this->method_description = 'Accept online payments on your ecommerce platform with a streamlined checkout your customers will love.'; // will be displayed on the options page

            $this->supports = array(
                'products',
                'refunds'
            );

            // Method with all the options fields
            $this->init_form_fields();

            // Load the settings.
            $this->init_settings();
            $this->title = $this->get_option('title');
            $this->description = $this->get_option('description');
            $this->enabled = $this->get_option('enabled');
            $this->store_name = $this->get_option('store_name');
            $this->dark_mode = $this->get_option('dark_mode');
            $this->show_logo = $this->get_option('show_logo');
            $this->widget_theme = $this->get_option('widget_theme');
            $this->is_inline_checkout = truevo_is_inline_checkout_flow();
            $this->debug_log = $this->get_option('debug_log');

            $this->testmode = truevo_is_test_mode_enabled() ? 'test' : 'live';
            $this->merchant_id = truevo_get_mid();
            $this->terminal_id = truevo_get_tid();
            $this->token_authorization = ($this->testmode == 'test') ? $this->get_option('test_token_authorization') : $this->get_option('token_authorization');

            $this->api_url = truevo_is_test_mode_enabled() ? 'https://sandbox.unify.truevo.com' : 'https://unify.truevo.com';

            $this->enable_authorization = $this->get_option('enable_authorization');
            $this->enable_tokenization = $this->get_option('enable_tokenization');
            
            $this->css_styles = $this->get_option('css_styles');

            $this->truevoapi = new Truevo_API_Wrapper($this->api_url, $this->token_authorization, $this->merchant_id, $this->terminal_id);

            $this->log = new WC_Logger();

            // This action hook saves the settings
            add_action('woocommerce_update_options_payment_gateways_' . $this->id, array($this, 'process_admin_options'));

            add_action('woocommerce_order_action_wc_truevo_order_action', array($this, 'truevo_wc_process_order_meta_box_action'));

            add_action('woocommerce_order_item_add_action_buttons', array($this, 'reverse_button'));

            add_action('woocommerce_receipt_' . $this->id, array($this, 'receipt_page'));

            add_action('woocommerce_api_' . $this->id, array($this, 'truevo_gateway_webhook'));

            // We need custom JavaScript to obtain a token
            add_action('wp_enqueue_scripts', array($this, 'payment_scripts'));

            if ($this->is_inline_checkout) {
                // FIXME: This is from WooCommerce 3.0+ and works slightly differently on older versions
                add_action('woocommerce_checkout_order_processed', array($this, 'truevo_order_processed'), 10, 2 );

            }
        }

        /**
         * Configuration fields in widget settings.
         *
         * @access public
         * @return void
         */
        public function init_form_fields()
        {
            $this->form_fields = array(
                'enabled' => array(
                    'title' => __('Enable Truevo Payments', 'truevo-gateway-woocommerce'),
                    'label' => __('Enable Truevo payments', 'truevo-gateway-woocommerce'),
                    'type' => 'select',
                    'options' => array(
                        'yes' => __('Yes', 'truevo-gateway-woocommerce'),
                        'no' => __('No', 'truevo-gateway-woocommerce')
                    ),
                    'default' => 'yes',
                    'description' => 'Start accepting payments.',
                ),
                'testmode' => array(
                    'title' => __('Enable Test Mode', 'truevo-gateway-woocommerce'),
                    'label' => __('Enable Test Mode', 'truevo-gateway-woocommerce'),
                    'type' => 'select',
                    'options' => array(
                        'yes' => __('Yes', 'truevo-gateway-woocommerce'),
                        'no' => __('No', 'truevo-gateway-woocommerce')
                    ),
                    'default' => 'no',
                    'description' => 'Test transactions before going live.',
                ),
                'test_merchant_id' => array(
                    'title' => __('Test Merchant ID (MID)', 'truevo-gateway-woocommerce'),
                    'type' => 'text',
                    'description' => __('If you require test credentials please contact Truevo support.', 'truevo-gateway-woocommerce'),
                    'desc_tip' => true,
                ),
                'test_terminal_id' => array(
                    'title' => __('Test Terminal ID (TID)', 'truevo-gateway-woocommerce'),
                    'type' => 'text',
                    'description' => __('If you require test credentials please contact Truevo support.', 'truevo-gateway-woocommerce'),
                    'desc_tip' => true,
                ),
                'live_merchant_id' => array(
                    'title' => __('Live Merchant ID (MID)', 'truevo-gateway-woocommerce'),
                    'type' => 'text',
                    'description' => __('Provided by Truevo once onboarding is complete.', 'truevo-gateway-woocommerce'),
                    'desc_tip' => true,
                ),
                'live_terminal_id' => array(
                    'title' => __('Live Terminal ID (TID)', 'truevo-gateway-woocommerce'),
                    'type' => 'text',
                    'description' => __('Provided by Truevo once onboarding is complete.', 'truevo-gateway-woocommerce'),
                    'desc_tip' => true,
                ),
                'token_authorization' => array(
                    'title' => __('Live Authentication Token', 'truevo-gateway-woocommerce'),
                    'type' => 'textarea',
                    'description' => __('Provided by Truevo once onboarding is complete.', 'truevo-gateway-woocommerce'),
                    'desc_tip' => true,
                ),
                'test_token_authorization' => array(
                    'title' => __('Test Authentication Token', 'truevo-gateway-woocommerce'),
                    'type' => 'textarea',
                    'description' => __('If you require test credentials please contact Truevo support.', 'truevo-gateway-woocommerce'),
                    'desc_tip' => true,
                ),
                'store_name' => array(
                    'title' => __('Store Name', 'truevo-gateway-woocommerce'),
                    'type' => 'text',
                    'description' => 'Add your store name to make reconciliation easier.',
                    'default' => '',
                ),
                'title' => array(
                    'title' => __('Payment Method Title', 'truevo-gateway-woocommerce'),
                    'type' => 'text',
                    'description' => __('What your customers sees when selecting a payment option.', 'truevo-gateway-woocommerce'),
                    'default' => __('', 'truevo-gateway-woocommerce'),

                ),
                'description' => array(
                    'title' => __('Payment Method Description', 'truevo-gateway-woocommerce'),
                    'type' => 'textarea',
                    'description' => __('Give your customers more information on how they can pay.', 'truevo-gateway-woocommerce'),
                    'default' => __('', 'truevo-gateway-woocommerce'),
                ),
                'enable_authorization' => array(
                    'title' => __('Enable Pre-Authorisation', 'truevo-gateway-woocommerce'),
                    'type' => 'select',
                    'options' => array(
                        'yes' => __('Yes', 'truevo-gateway-woocommerce'),
                        'no' => __('No', 'truevo-gateway-woocommerce')
                    ),
                    'default' => 'yes',
                    'description' => __('Only available if pre-authorisation is set up on your Truevo account.', 'truevo-gateway-woocommerce'),
                    'desc_tip' => true,
                    'custom_attributes' => [
                        'data-description' => __( 'Reserve funds and capture the payment later.', 'truevo-gateway-woocommerce' ),
                    ],
                ),
                'enable_tokenization' => array(
                    'title' => __('Enable Saved Cards', 'truevo-gateway-woocommerce'),
                    'type' => 'select',
                    'options' => array(
                        'yes' => __('Yes', 'truevo-gateway-woocommerce'),
                        'no' => __('No', 'truevo-gateway-woocommerce')
                    ),
                    'default' => 'yes',
                    'description' => __('Your customers can save their card details for repeat purchases.', 'truevo-gateway-woocommerce'),
                ),
                'checkout_flow' => array(
                    'title' => __('Checkout Flow', 'truevo-gateway-woocommerce'),
                    'type' => 'select',
                    'options' => array(
                        'inline' => __('Inline', 'truevo-gateway-woocommerce'),
                        'redirect' => __('Redirect', 'truevo-gateway-woocommerce')
                    ),
                    'default' => 'inline',
                ),
                'widget_theme' => array(
                    'title' => __('Plugin Theme', 'truevo-gateway-woocommerce'),
                    'type' => 'select',
                    'options' => array(
                        'classic' => __('Classic', 'truevo-gateway-woocommerce'),
                        'minimalist' => __('Minimalist', 'truevo-gateway-woocommerce'),
                        'expanded' => __('Expanded', 'truevo-gateway-woocommerce'),
                        'contained' => __('Contained', 'truevo-gateway-woocommerce'),
                        'compact' => __('Compact', 'truevo-gateway-woocommerce')
                    ),
                    'default' => 'classic',
                ),
                'dark_mode' => array(
                    'title' => __('Dark Mode', 'truevo-gateway-woocommerce'),
                    'type' => 'checkbox',
                    'default' => 'no',
                    'description' => __('Toggle this to change your payment form into dark mode. This will not change your whole site to dark mode.'),
                    'desc_tip' => true,
                     'custom_attributes' => [
                       'toggle-switch' 
                 ],
                ),
                'show_logo' => array(
                    'title' => __('Show Security Statement', 'truevo-gateway-woocommerce'),
                    'type' => 'select',
                    'options' => array(
                        'yes' => __('Yes', 'truevo-gateway-woocommerce'),
                        'no' => __('No', 'truevo-gateway-woocommerce')
                    ),
                    'default' => 'no',
                    'description' => __('Show the Truevo security statement, so your customers know it&#39;s safe to make payment on your store.', 'truevo-gateway-woocommerce'),
                    'desc_tip' => true,
                     'custom_attributes' => [
                       'data-additional-description' => __( 'Show the Truevo security statement, so your customers know it&#39;s safe to make payment on your store.', 'truevo-gateway-woocommerce' ),
                 ],
                ),
                'css_styles' => array(
                    'title' => __('Custom CSS', 'truevo-gateway-woocommerce'),
                    'type' => 'textarea',
                    'description' => __('Customise colours and fonts to match your website.', 'truevo-gateway-woocommerce'),
                    'default' => '',
                ),
                'debug_log' => array(
                    'title' => __('Enable Debug Log', 'truevo-gateway-woocommerce'),
                    'type' => 'checkbox',
                    'label' => __('This will log the gateway&#39;s requests and responses which will be stored in:', 'truevo-gateway-woocommerce'),
                    'default' => 'no',
                    'description' => sprintf( '<code>' . WC_Log_Handler_File::get_log_file_path('truevo_gateway') . '</code>')
                )
            );
        }
   
        public function get_option( $key, $empty_value = null ) {
            $setting = parent::get_option($key, $empty_value);
            
            if ($key === 'css_styles') {
                return html_entity_decode($setting);
            }
            
            return $setting;
        }

        /**
         * Generate the configuration and display the payment widget on the page.
         * The "redirect" checkout flow will have an order to build the configuration with.
         * The "inline" checkout flow will need to make use of the cart / session.
         *
         * @param WC_Order|null $order
         * @return void
         */
        public function display_payment_widget(WC_Order $order = null)
        {
            // Get the widget configuration
            $truevo_widget_configuration_builder = new Truevo_Widget_Configuration_Builder($order);
            $configuration = $truevo_widget_configuration_builder->build();

            // Order / Cart Identifier
            $identifier = $truevo_widget_configuration_builder->getOrderIdentifierAttribute();

            // Generate URL to POST transaction response to
            $url_user_on_success = !is_null($order) ? truevo_get_transaction_endpoint($identifier) : null;

            ob_start();

            include_once 'templates/widget.php';

            echo ob_get_clean();

            // Enqueue scripts for the widget on checkout
            wp_register_script( 'wc_truevo_checkout_widget', plugins_url( 'assets/js/truevo-checkout-widget.js', __FILE__ ), ['jquery', 'wc_truevo_payment_request'] );
            wp_localize_script(
                'wc_truevo_checkout_widget',
                'wc_truevo_checkout_widget_params',
                array(
                    'identifier' => !is_null($order) ? "Order #$identifier" : "Inline Checkout $identifier",
                    'configuration' => $configuration,
                    'is_inline' => $this->is_inline_checkout
                )
            );
            wp_enqueue_script( 'wc_truevo_checkout_widget' );
        }

        /**
         * Display the plugin description on the checkout page.
         *
         * @access public
         * @return void
         */
        public function payment_fields()
        {
            // Ok, let's display some description before the payment form
            if ($this->description) {
                // If test mode is enabled, display example test card number.
                if ($this->testmode == 'test') {
                    $this->description .= wpautop('&nbsp;');
                    $this->description .= wpautop('TEST MODE ENABLED. <br> <br> In test mode you can use the card numbers listed in our <a href="https://docs.truevo.com/payment-plugins/woocommerce/" target="_blank" rel="noopener noreferrer">Developer Docs</a>. Remember to disable test mode before going live.');
                    $this->description = trim($this->description);
                }
                // Display the description with <p> tags etc.
                echo wpautop(wp_kses_post($this->description));
            }

            // Display the payment widget on the checkout page
            if ($this->is_inline_checkout) {
                $this->display_payment_widget();
            }
        }

        /**
         * Load the required scripts needed for the payment widget.
         *
         * @access public
         * @return void
         */
        public function payment_scripts()
        {
            wp_enqueue_script("woocommerce-truevo-inline-widget", plugins_url('assets/js/truevo-inline-widget.js', __FILE__), [], strtotime('now'), true);

            // we need JavaScript to process a token only on cart/checkout pages, right?
            if (!is_cart() && !is_checkout() && !isset($_GET['pay_for_order'])) {
                return;
            }

            // if our payment gateway is disabled, we do not have to enqueue JS too
            if ('no' === $this->enabled) {
                return;
            }

            // no reason to enqueue JavaScript if API keys are not set
            if (empty($this->merchant_id) || empty($this->terminal_id)) {
                return;
            }

            wp_enqueue_style("woocommerce-truevo-widget", plugins_url('assets/css/truevo-widget.css', __FILE__), [], strtotime('now'));
            wp_enqueue_script("woocommerce-truevo-apple-pay", "https://applepay.cdn-apple.com/jsapi/v1/apple-pay-sdk.js");

            if ($this->testmode === 'live') {
                wp_enqueue_script("woocommerce-truevo-merchant", "https://pay.truevo.com/widget/merchant.minified.js", [], null, true);
            } else {
                wp_enqueue_script("woocommerce-truevo-merchant", "https://sandbox-truevo-drop-in-hosted-payment-widget.s3.eu-west-1.amazonaws.com/merchant.minified.js", [], null, true);
            }

            // Register AJAX handler for logging
            wp_register_script( 'wc_truevo_payment_request', plugins_url( 'assets/js/jquery.ajax.js', __FILE__ ), ['jquery']);
            wp_localize_script(
                'wc_truevo_payment_request',
                'wc_truevo_payment_request_params',
                array(
                    'ajax_url' => WC_AJAX::get_endpoint( 'wc_truevo_add_message_to_log' ),
                    'security' => wp_create_nonce('truevo-ajax-nonce'),
                    'url_add_message_to_log' => WC_AJAX::get_endpoint( 'wc_truevo_add_message_to_log' ),
                )
            );
            wp_enqueue_script( 'wc_truevo_payment_request' );
        }

        /**
         * Process an order for the inline checkout flow and add intent to order notes.
         *
         * @param $order_id
         * @param $posted_data
         * @return void
         */
        public function truevo_order_processed($order_id, $posted_data)
        {
            $order = wc_get_order( $order_id );

            $payment_method = wc_clean( wp_unslash( $_POST['payment_method'] ) );

            if ( $payment_method === $this->id ) {
                $order->add_order_note(
                    sprintf(
                        'Inline Truevo Checkout Payment intent created (%1$s)',
                        WC()->cart->get_cart_hash()
                    )
                );
            }
        }

        /**
         * The receipt page which loads the widget to accept payments in the "redirect" checkout flow.
         *
         * @access public
         * @return void
         */
        public function receipt_page($order)
        {
            if ($this->debug_log == 'yes') {
                $this->log->add('truevo_gateway', 'Truevo order pay page loading for Order #' . $order);
            }

            $order = wc_get_order($order);

            if (!$this->is_inline_checkout) {
                $this->display_payment_widget($order);
            }
        }

        /**
         * Endpoint to POST response from the payment widget.
         *
         * @access public
         * @return void
         */
        public function truevo_gateway_webhook()
        {
            if (!check_ajax_referer('truevo-security-nonce', 'security', false)) {
                wp_send_json_error('Invalid security token sent.');
                wp_die();
            }

            $order_id = sanitize_key($_GET['response_id']);
            $response_secret = sanitize_text_field($_GET['response_secret']);
            $order = wc_get_order($order_id);

            $truevo_order_flag = isset($_POST['truevo-order-flag']) ? $_POST['truevo-order-flag'] : '';
            $truevo_order_values = isset($_POST['truevo-order-values']) ? $_POST['truevo-order-values'] : '';
            $truevo_order_values_all = isset($_POST['truevo-order-values-all']) ? $_POST['truevo-order-values-all'] : '';

            if ($this->debug_log == 'yes') {
                $truevo_order_values_all = stripslashes($truevo_order_values_all);
                $truevo_order_values_all = json_decode($truevo_order_values_all, true);

                $this->log->add('truevo_gateway', 'Truevo gateway transaction response for Order #' . $order_id . '. ' . print_r($truevo_order_values_all, true));
            }

            if (md5("response_id" . $order_id) == $response_secret) {
                $order_completed = false;

                $truevo_order_values = stripslashes($truevo_order_values);
                $truevo_order_values = json_decode($truevo_order_values, true);

                foreach ($truevo_order_values as $key => $truevo_order_value) {
                    $order->update_meta_data('truevo_payment_' . $key, $truevo_order_value);
                }

                $redirect_url = $order->get_checkout_payment_url();

                if ($truevo_order_flag == "declined") {
                    $this->add_order_note_with_transaction_details($order, 'Truevo Payments: The payment was declined.', $truevo_order_values);

                    wc_add_notice(__('Truevo Payments: Your payment failed. Please try again or use a different card.', 'truevo-gateway-woocommerce'), 'error');
                }

                if ($truevo_order_flag == "failed") {
                    $this->add_order_note_with_transaction_details($order, 'Truevo Payments: Payment Failed.', $truevo_order_values);

                    wc_add_notice(__('Truevo Payments: Your payment failed. Please try again or use a different card.', 'truevo-gateway-woocommerce'), 'error');
                }

                if ($truevo_order_flag == "completed") {

                    $truevo_gateway_settings = get_option('woocommerce_truevo_gateway_settings');

                    if ($truevo_gateway_settings['enable_authorization'] == 'yes') {
                        $order->update_status('on-hold');

                        $order->update_meta_data('truevo_payment_transaction_type', 'authorize');

                        // Add hidden meta data for getting amount to capture (starting with "_" will hide the meta data on the edit order page)
                        $order->update_meta_data('_truevo_payment_preauth_currency', $order->get_currency());
                        $order->update_meta_data('_truevo_payment_preauth_amount', $order->get_total());

                        $this->add_order_note_with_transaction_details($order, 'The payment was pre-authorized. Please capture the payment when ready.', $truevo_order_values);
                    } else {
                        $order->update_status('processing');
                        $order->reduce_order_stock();

                        $order->update_meta_data('truevo_payment_transaction_type', 'sale');
                        $this->add_order_note_with_transaction_details($order, 'Truevo Payments: The payment was successfully completed.', $truevo_order_values);
                    }

                    $redirect_url = $this->get_return_url($order);

                    $order_completed = true;
                }

                if ($order->save()) {
                    if ($order_completed) {
                        WC()->cart->empty_cart();

                        if ($this->is_inline_checkout) {
                            wp_send_json([
                                'redirect_url' => $this->get_return_url($order)
                            ]);
                        }
                    }
                }

                if (!$this->is_inline_checkout) {
                    wp_redirect($redirect_url);
                }
            }
        }

        /**
         * Process the payment when the capture method is selected on the backend order page.
         *
         * @param \WC_Order $order
         */
        public function truevo_wc_process_order_meta_box_action($order)
        {
            // Get Order ID
            $order_id = $order->id;

            // Get Post Meta
            $truevo_payment_id = get_post_meta($order_id, 'truevo_payment_transactionId', true);
            $truevo_payment_rrn = get_post_meta($order_id, 'truevo_payment_rrn', true);
            $truevo_payment_captured = get_post_meta($order_id, 'truevo_payment_captured', true);
            $truevo_payment_type = get_post_meta($order_id, 'truevo_payment_transaction_type', true);

            // Check if payment details are available and if it is an authorized payment that is not yet captured
            if ( !empty($truevo_payment_id) && $truevo_payment_rrn && $truevo_payment_captured != 'yes' && $truevo_payment_type == 'authorize' ) {
                // Setup API request
                $data_info = array(
                    'transactionUniqueIdentifier' => $truevo_payment_id,
                    'retrievalReferenceNumber' => $truevo_payment_rrn,
                    'amount' => number_format($order->get_total(), 2, '.', ''),
                    'merchantReference' => (string)$order->get_id(),
                    'paymentReference' => $this->store_name,
                    'billing_address' => $order->get_billing_address_1(),
                    'billing_address_2' => $order->get_billing_address_2(),
                    'billing_city' => $order->get_billing_city(),
                    'billing_state' => $order->get_billing_state(),
                    'billing_country' => $order->get_billing_country(),
                    'billing_postcode' => $order->get_billing_postcode(),
                    'billing_first_name' => $order->billing_first_name,
                    'billing_last_name' => $order->billing_last_name,
                    'delivery_address' => $order->get_shipping_address_1(),
                    'delivery_address_2' => $order->get_shipping_address_2(),
                    'delivery_city' => $order->get_shipping_city(),
                    'delivery_country' => $order->get_shipping_country(),
                    'delivery_postcode' => $order->get_shipping_postcode(),
                    'delivery_state' => $order->get_shipping_state(),
                    'email' => $order->get_billing_email(),
                    'user_ip' => truevo_get_client_ip(),
                );

                // Log request
                if ($this->debug_log == 'yes') {
                    $this->log->add('truevo_gateway', 'Truevo payments capture request. ' . print_r($data_info, true));
                }

                // Send request
                $process_capture = $this->truevoapi->process_truevo_capture($data_info);

                // Log initial response
                if ($this->debug_log == 'yes') {
                    $this->log->add('truevo_gateway', 'Truevo payments capture response. ' . print_r($process_capture, true));
                }

                $rejected = true;

                // Check if the capture request was received
                if ($process_capture['code'] == '200' && !empty($process_capture['result'])) {
                    if ($process_capture['result']->outcome->status === 'RECEIVED') {

                        // Poll the status of the capture request using the Transaction ID
                        $transaction_id = $process_capture['result']->transactionUniqueIdentifier;
                        $query_final_response = $this->truevoapi->poll_transaction_status($transaction_id, $this->debug_log);

                        // Get the final response
                        $query_final_status = $query_final_response['status'] ?? '';
                        $query_final_message = $query_final_response['reason'] ?? '';

                        // Update post meta and add order note
                        if ($query_final_status === 'COMPLETE') {
                            // Get new Transaction ID
                            $transaction_id = $process_capture['result']->transactionUniqueIdentifier;

                            $order->update_status('processing');
                            $this->add_order_note_with_transaction_details($order, 'Truevo Payments: Your payment was successfully captured.', $query_final_response['additionalInfo']);

                            update_post_meta($order_id, 'truevo_payment_captured', 'yes');
                            update_post_meta($order_id, 'truevo_payment_captured_status', $query_final_status);
                            update_post_meta($order_id, 'truevo_payment_transactionId', $transaction_id);

                            $rejected = false;
                        } else {
                            $this->add_order_note_with_transaction_details($order, $query_final_message, $query_final_response['additionalInfo']);
                            update_post_meta($order_id, 'truevo_payment_captured_status', $query_final_status);
                        }
                    }
                }

                // Add order note if rejected and update status
                if ($rejected) {
                    $order->update_status('on-hold');
                    $order->add_order_note('Truevo Payments: Payment capture could not be processed.', false);
                }

                // Save the updated order
                $order->save();
            }
        }

        /**
         * Process the payment field and redirect to checkout/pay page.
         *
         * @access public
         * @param mixed $order_id
         * @return array
         */
        public function process_payment($order_id)
        {
            $order = wc_get_order($order_id);

            $truevo_widget_configuration_builder = new Truevo_Widget_Configuration_Builder($order);

            // If inline checkout is enabled, a fake error needs to be generated to validate the order before accepting payment
            if ($this->is_inline_checkout) {
                if ($_POST['validate_order_inline_checkout'] == "1") {
                    wc_add_notice( __( "Submitting transaction", 'truevo-gateway-woocommerce' ), 'notice', array(
                        'truevo-inline-flow-validated' => true,
                        'truevo-order-transaction' => json_encode($truevo_widget_configuration_builder->getTransactionAttribute()),
                        'truevo-order-customer' => json_encode($truevo_widget_configuration_builder->getCustomerAttribute()),
                        'truevo-submit-url' => truevo_get_transaction_endpoint($order_id)
                     ));

                    return array(
                        'result' => 'fail',
                        'redirect' => '',
                    );
                }

                return [
                    'result'   => 'success',
                    'redirect' => $this->get_return_url($order)
                ];
            }

            return array(
                'result' => 'success',
                'redirect' => $order->get_checkout_payment_url(true)
            );
        }

        /**
         * Process the reversal/refund when the corresponding action button is clicked on the backend order page.
         *
         * @access public
         *
         * @param mixed $order_id
         * @param mixed $amount (default: null)
         * @param string $reason (default: '')
         *
         * @return bool
         */
        public function process_refund($order_id, $amount = null, $reason = '')
        {
            // Setup refund/reverse transaction
            $refund_mode = $reason === 'Reverse_Full_Payment' ? 'reverse' : 'refund';

            $order = wc_get_order($order_id);
            $truevo_payment_id = get_post_meta($order_id, 'truevo_payment_transactionId', true);
            $transactionType = get_post_meta($order_id, 'truevo_payment_transaction_type', true);
            if ( !empty($truevo_payment_id) ) {
                $data_info = array(
                    'transactionUniqueIdentifier' => $truevo_payment_id,
                    'transactionType' => strtolower($transactionType) === "authorize" ? "capture" : "sale",
                    'amount' => $amount,
                    'merchantReference' => (string) $order->get_id(),
                    'paymentReference' => $this->store_name,
                    'billing_address' => $order->get_billing_address_1(),
                    'billing_address_2' => $order->get_billing_address_2(),
                    'billing_city' => $order->get_billing_city(),
                    'billing_state' => $order->get_billing_state(),
                    'billing_country' => $order->get_billing_country(),
                    'billing_postcode' => $order->get_billing_postcode(),
                    'billing_first_name' => $order->billing_first_name,
                    'billing_last_name' => $order->billing_last_name,
                    'delivery_address' => $order->get_shipping_address_1(),
                    'delivery_address_2' => $order->get_shipping_address_2(),
                    'delivery_city' => $order->get_shipping_city(),
                    'delivery_country' => $order->get_shipping_country(),
                    'delivery_postcode' => $order->get_shipping_postcode(),
                    'delivery_state' => $order->get_shipping_state(),
                    'email' => $order->get_billing_email(),
                    'user_ip' => truevo_get_client_ip(),
                );

                // Log request
                if ($this->debug_log == 'yes') {
                    $this->log->add('truevo_gateway', 'Truevo payments refund/reverse request. ' . print_r($data_info, true));
                }

                // Send request
                $process_refund = $this->truevoapi->process_truevo_refund($data_info, $refund_mode);

                // Log initial response
                if ($this->debug_log == 'yes') {
                    $this->log->add('truevo_gateway', 'Truevo payments refund response. ' . print_r($process_refund, true));
                }

                $rejected = true;

                // Check if the refund/reversal request was received
                if ( $process_refund['code'] == '200' && !empty($process_refund['result']) ) {
                    if ( $process_refund['result']->outcome->status === 'RECEIVED' ) {

                        // Poll the status of the refund/reversal request using the Transaction ID
                        $transaction_id = $process_refund['result']->transactionUniqueIdentifier;
                        $refund_final_response = $this->truevoapi->poll_transaction_status($transaction_id, $this->debug_log);

                        // Get the final response
                        $refund_final_status = $refund_final_response['status'] ?? '';
                        $refund_final_message = $refund_final_response['reason'] ?? '';

                        // Log the final response
                        if ($this->debug_log == 'yes') {
                            $this->log->add('truevo_gateway', 'Truevo payments refund status : ' . $refund_final_status);
                        }

                        // Update post meta and add order note
                        if ( $refund_final_status === 'COMPLETE' ) {
                            update_post_meta($order_id, 'truevo_payment_refund_status', $refund_final_status);

                            $rejected = false;

                            $order->add_order_note(sprintf('Truevo Payments : The payment %s was successfully refunded.', $amount), false);
                        } else {
                            update_post_meta($order_id, 'truevo_payment_refund_status', $refund_final_status);

                            if ($refund_final_message) {
                                $order->add_order_note($refund_final_message, false);
                            }
                        }
                    }
                }

                // Save the updated order
                $order->save();

                // Return true or false depending on success status
                if ($rejected) {
                    $order->add_order_note(sprintf('Truevo Payments: The refund was declined.', $amount), false);
                    return false;
                } else {
                    return true;
                }
            }

            // Failed to get response from the Truevo API
            $order->add_order_note("Truevo Payments: The refund could not be processed.", false);
            return false;
        }

        /**
         * Display the reverse button on the backend order page.
         *
         * @param $order
         * @return void
         */
        public function reverse_button($order)
        {
            ?>
            <button type="button" class="button reverse-items"
                    data-order_id="<?php echo $order->get_id(); ?>"><?php esc_html_e('Reverse', 'woocommerce'); ?></button>
            <?php
        }

        /**
         * Add order note that includes required transaction details if they are set.
         *
         * @param $order
         * @param $order_note
         * @param $order_values
         * @return void
         */
        private function add_order_note_with_transaction_details($order, $order_note, $order_values)
        {
            // RRN
            $order_note .= "\nRRN: ";
            $order_note .= isset($order_values['rrn']) ? $order_values['rrn'] : '-';

            // Response Code
            $order_note .= "\nResponse Code: ";
            $order_note .= isset($order_values['responseCode']) ? $order_values['responseCode'] : '-';

            // Response Code
            $order_note .= "\nResponse Description: ";
            $order_note .= isset($order_values['reasonDescription']) ? $order_values['reasonDescription'] : '-';

            // Transation Status
            $order_note .= "\nTransaction Status: ";
            $order_note .= isset($order_values['transactionStatus']) ? $order_values['transactionStatus'] : '-';

            $order->add_order_note($order_note, false);
        }
    }
}

/**
 * Get the Truevo API Wrapper and populate it with the correct configuration
 *
 * @return Truevo_API_Wrapper|false
 */
function truevo_api_connection()
{
    $truevo_gateway_settings = get_option('woocommerce_truevo_gateway_settings');
    if (!empty($truevo_gateway_settings)) {
        $testmode = ('yes' === $truevo_gateway_settings['testmode']) ? 'test' : 'live';
        $merchant_id = ($testmode == 'test') ? $truevo_gateway_settings['test_merchant_id'] : $truevo_gateway_settings['live_merchant_id'];
        $terminal_id = ($testmode == 'test') ? $truevo_gateway_settings['test_terminal_id'] : $truevo_gateway_settings['live_terminal_id'];
        $token_authorization = ($testmode == 'test') ? $truevo_gateway_settings['test_token_authorization'] : $truevo_gateway_settings['token_authorization'];

        $api_url = ($testmode == 'test') ? 'https://sandbox.unify.truevo.com' : 'https://unify.truevo.com';

        $truevoapi = new Truevo_API_Wrapper($api_url, $token_authorization, $merchant_id, $terminal_id);

        return $truevoapi;
    }

    return false;
}

/**
 * Get the current client's IP address (not including localhost).
 *
 * @access public
 * @return string|null
 */
function truevo_get_client_ip()
{
    $ipAddress = isset($_SERVER['REMOTE_ADDR']) ? trim($_SERVER['REMOTE_ADDR']) : null;

    // Check IP address is valid and not localhost
    if (filter_var($ipAddress, FILTER_VALIDATE_IP) && (substr($ipAddress, 0, 4) != '127.' && $ipAddress != '::1')) {
        return $ipAddress;
    }

    return null;
}

/**
 * Generate a unique merchant shopper reference for a logged in
 * customer to use for tokenization.
 *
 * @return string|null
 */
function truevo_get_merchant_shopper_reference()
{
    if ( is_user_logged_in() ) {
        $customer = wp_get_current_user();
        return "tokenization-" . $customer->ID . '-' . $customer->user_email . '-' . strtotime($customer->user_registered);
    }

    return null;
}

/**
 * Is the inline checkout flow enabled?
 *
 * @return bool
 */
function truevo_is_inline_checkout_flow()
{
    $truevo_gateway_settings = get_option('woocommerce_truevo_gateway_settings');

    return $truevo_gateway_settings['checkout_flow'] !== 'redirect';
}

/**
 * Is the sandbox / test mode enabled?
 *
 * @return bool
 */
function truevo_is_test_mode_enabled()
{
    $truevo_gateway_settings = get_option('woocommerce_truevo_gateway_settings');

    return $truevo_gateway_settings['testmode'] === 'yes';
}

/**
 * Get the merchant identifier based on if test mode is enabled or not.
 *
 * @return string
 */
function truevo_get_mid()
{
    $truevo_gateway_settings = get_option('woocommerce_truevo_gateway_settings');

    return truevo_is_test_mode_enabled() ? $truevo_gateway_settings['test_merchant_id'] : $truevo_gateway_settings['live_merchant_id'];
}

/**
 * Get the terminal identifier based on if test mode is enabled or not.
 *
 * @return string
 */
function truevo_get_tid()
{
    $truevo_gateway_settings = get_option('woocommerce_truevo_gateway_settings');

    return truevo_is_test_mode_enabled() ? $truevo_gateway_settings['test_terminal_id'] : $truevo_gateway_settings['live_terminal_id'];
}

/**
 * Get the endpoint to POST transaction response to.
 *
 * @param $order_id
 * @return mixed
 */
function truevo_get_transaction_endpoint($order_id)
{
    return add_query_arg(
        array(
            'response_id' => $order_id,
            'response_secret' => md5("response_id" . $order_id)
        ),
        site_url('/wc-api/truevo_gateway')
    );
}

// Add actions
add_action('woocommerce_order_actions', 'truevo_wc_add_order_meta_box_action');
add_action('admin_head', 'truevo_hide_show_wc_refund_reverse_buttons');
add_action('admin_enqueue_scripts', 'truevo_backend_scripts');
add_action('wp_ajax_nopriv_truevo_gateway_reverse_items', 'truevo_gateway_reverse_items');
add_action('wp_ajax_truevo_gateway_reverse_items', 'truevo_gateway_reverse_items');
add_action('wc_ajax_wc_truevo_add_message_to_log', 'truevo_ajax_add_message_to_log');

/**
 * Add a custom action to order actions select box on edit order page.
 * Only added for capture payments.
 *
 * @param array $actions order actions array to display
 * @return array - updated actions
 */
function truevo_wc_add_order_meta_box_action($actions)
{
    global $theorder;

    if ($theorder->get_payment_method() !== 'truevo_gateway') {
        return;
    }

    $truevo_payment_captured = get_post_meta($theorder->id, 'truevo_payment_captured', true);
    $truevo_payment_id = get_post_meta($theorder->id, 'truevo_payment_transactionId', true);
    $truevo_payment_type = get_post_meta($theorder->id, 'truevo_payment_transaction_type', true);

    if ($truevo_payment_captured != 'yes' && $truevo_payment_type == 'authorize') {
        $truevoapi = truevo_api_connection();
        $get_payment_details = $truevoapi->get_payment_details($truevo_payment_id, true);

        if ($get_payment_details === 'capture') {
            $actions['wc_truevo_order_action'] = __('Capture the payment', 'truevo-gateway-woocommerce');
        }
    }

    return $actions;
}

/**
 * Check conditions of transaction and hide the refund and reversal action buttons where applicable.
 *
 * @return void
 */
function truevo_hide_show_wc_refund_reverse_buttons()
{

    global $post;
    if (strpos($_SERVER['REQUEST_URI'], 'post.php?post=') === false) {
        return;
    }

    if (empty($post) || $post->post_type != 'shop_order') {
        return;
    }

    $order_id = $post->ID;

    $order = wc_get_order($order_id);

    if ($order->get_payment_method() !== 'truevo_gateway') {
        return;
    }

    $truevo_payment_id = get_post_meta($order_id, 'truevo_payment_transactionId', true);
    $truevo_gateway_settings = get_option('woocommerce_truevo_gateway_settings');

    if ($truevo_payment_id) {

        $truevoapi = truevo_api_connection();
        $refund_mode = $truevoapi->get_payment_details( $truevo_payment_id, false, $truevo_gateway_settings['debug_log'] );
        if ($refund_mode == "reverse") {
            ?>
            <style type="text/css">
                .refund-items {
                    display: none !important
                }
            </style>
            <?php
        } else if ($refund_mode == "refund") {
            ?>
            <style type="text/css">
                .reverse-items {
                    display: none !important
                }
            </style>
            <?php
        } else {
            ?>
            <style type="text/css">
                .refund-items, .reverse-items {
                    display: none !important
                }
            </style>
            <?php
        }
    }

}

/**
 * Add required scripts for thr Truevo Admin section.
 *
 * @access public
 * @return void
 */
function truevo_backend_scripts()
{   
    wp_register_script('woocommerce-truevo-admin', plugins_url('assets/js/truevo-widget-admin.js', __FILE__), array('jquery'), strtotime('now'));
    wp_enqueue_style("woocommerce-truevo-widget", plugins_url('assets/css/truevo-settings.css', __FILE__), [], strtotime('now'), 'all');
    
    wp_localize_script(
        'woocommerce-truevo-admin',
        'truevo_ajax_object',
        [
            'security' => wp_create_nonce('truevo-admin-nonce'),
        ]
    );
    wp_enqueue_script('woocommerce-truevo-admin');
}

/**
 * Allow for reversals (same day full refund) on the admin order page.
 *
 * @access public
 * @return void
 */
function truevo_gateway_reverse_items()
{

    if (!check_ajax_referer('truevo-admin-nonce', 'security', false)) {

        wp_send_json_error('Invalid security token sent.');
        wp_die();
    }


    if (isset($_POST['order_id']) && $_POST['order_id'] != "") {

        $log = new WC_Logger();

        $order_id = $_POST['order_id'];
        $order = wc_get_order($order_id);

        if (!is_a($order, 'WC_Order')) {
            return;
        }

        $truevo_gateway_settings = get_option('woocommerce_truevo_gateway_settings');


        if (!empty($truevo_gateway_settings)) {

            // Get Items
            $order_items = $order->get_items();

            // Refund Amount
            $refund_amount = 0;

            // Prepare line items which we are refunding
            $line_items = array();

            if ($order_items = $order->get_items()) {

                foreach ($order_items as $item_id => $item) {

                    $item_meta = $order->get_item_meta($item_id);

                    $product_data = wc_get_product($item_meta["_product_id"][0]);

                    $item_ids[] = $item_id;
                    $tax_data = $item_meta['_line_tax_data'];
                    $refund_tax = 0;

                    if (is_array($tax_data[0])) {

                        $refund_tax = array_map('wc_format_decimal', $tax_data[0]);

                    }

                    // FIXME: Refund amount is not including shipping cost
                    $refund_amount = wc_format_decimal($refund_amount) + wc_format_decimal($item_meta['_line_total'][0]);

                    $line_items[$item_id] = array('qty' => $item_meta['_qty'][0], 'refund_total' => wc_format_decimal($item_meta['_line_total'][0]), 'refund_tax' => $refund_tax);
                }
            }

            $refund = wc_create_refund(array(
                'amount' => $refund_amount,
                'reason' => "Reverse_Full_Payment",
                'order_id' => $order_id,
                'line_items' => $line_items,
                'refund_payment' => true
            ));

            if (is_wp_error($refund)) {
                $error_string = $refund->get_error_message();
                wp_send_json_error(json_encode($error_string));
            } else {
                if ($truevo_gateway_settings['debug_log'] == 'yes') {
                    $log->add('truevo_gateway', 'Truevo reverse processed. ' . print_r($refund, true));
                }
                wp_send_json_success();
            }
        }
    }

    wp_die();
}

/**
 * Endpoint specific to logging frontend actions.
 *
 * @return Response
 */
function truevo_ajax_add_message_to_log()
{
    check_ajax_referer( 'truevo-ajax-nonce', 'security' );

    $log = new WC_Logger();
    $truevo_gateway_settings = get_option('woocommerce_truevo_gateway_settings');

    if ($truevo_gateway_settings['debug_log'] == 'yes') {
        $message = sanitize_text_field( $_POST['message'] );

        $details = $_POST['details'];

        if (isset($details) && is_array($details) && !empty($details)) {
            $message .= ' ' . sanitize_textarea_field(print_r($details, true));
        }

        $log->add('truevo_gateway', $message);
    }

    wp_send_json([
        'success' => true
    ]);
}